import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:adeptutils/adeptutils.dart';
import 'dart:ui' as ui show BoxHeightStyle, BoxWidthStyle;

enum CustomButton {
  NONE,
  ALWAYS,
  ON_TEXT,
}

enum CustomButtonDefaultAction {
  NONE,
  CLEAR,
  PASSWORD_TOGGLE,
}

// ignore_for_file: annotate_overrides, overridden_fields
class AdeptTextField extends StatefulWidget {
  static Widget _defaultContextMenuBuilder(
      BuildContext context, EditableTextState editableTextState) {
    return AdaptiveTextSelectionToolbar.editableText(
      editableTextState: editableTextState,
    );
  }

  final Key? key;
  final AdeptTextEditingController? controller;
  final FocusNode? focusNode;
  final InputDecoration decoration;
  final TextInputType? keyboardType;
  final TextInputAction? textInputAction;
  final TextCapitalization textCapitalization;
  final TextStyle? style;
  final TextAlign textAlign;
  final TextDirection? textDirection;
  final bool autofocus;
  final bool obscureText;
  final bool autocorrect;
  final int maxLines;
  final int? maxLength;
  final MaxLengthEnforcement? maxLengthEnforcement;
  final ValueChanged<String>? onChanged;
  final VoidCallback? onEditingComplete;
  final ValueChanged<String>? onSubmitted;
  final List<TextInputFormatter>? inputFormatters;
  final bool? enabled;
  final double cursorWidth;
  final Radius? cursorRadius;
  final Color? cursorColor;
  final Brightness? keyboardAppearance;
  final EdgeInsets scrollPadding;
  final bool enableInteractiveSelection;
  final TextSelectionControls? selectionControls;
  final GestureTapCallback? onTap;
  final StrutStyle? strutStyle;
  final TextAlignVertical? textAlignVertical;
  final bool readOnly;
  final bool? showCursor;
  final String obscuringCharacter;
  final SmartDashesType? smartDashesType;
  final SmartQuotesType? smartQuotesType;
  final bool enableSuggestions;
  final int? minLines;
  final bool expands;
  final AppPrivateCommandCallback? onAppPrivateCommand;
  final double? cursorHeight;
  final ui.BoxHeightStyle selectionHeightStyle;
  final ui.BoxWidthStyle selectionWidthStyle;
  final DragStartBehavior dragStartBehavior;
  final MouseCursor? mouseCursor;
  final InputCounterWidgetBuilder? buildCounter;
  final ScrollController? scrollController;
  final ScrollPhysics? scrollPhysics;
  final Iterable<String>? autofillHints;
  final String? restorationId;
  final TapRegionCallback? onTapOutside;
  final Clip clipBehavior;
  final bool scribbleEnabled;
  final bool enableIMEPersonalizedLearning;
  final EditableTextContextMenuBuilder? contextMenuBuilder;
  final SpellCheckConfiguration? spellCheckConfiguration;
  final TextMagnifierConfiguration? magnifierConfiguration;

  final CustomButton customButton;
  final CustomButtonDefaultAction customButtonDefaultAction;
  final VoidCallback? customButtonAction;
  final Color? customButtonColor;
  final IconData? customButtonIcon;
  final double customButtonIconSize;
  final double customButtonIconPadding;
  final bool clearErrorOnTextChange;

  const AdeptTextField({
    this.key,
    this.controller,
    this.focusNode,
    InputDecoration? decoration,
    this.keyboardType,
    this.textInputAction,
    TextCapitalization? textCapitalization,
    this.style,
    TextAlign? textAlign,
    this.textDirection,
    bool? autofocus,
    bool? obscureText,
    bool? autocorrect,
    int? maxLines,
    this.maxLength,
    this.maxLengthEnforcement,
    this.onChanged,
    this.onEditingComplete,
    this.onSubmitted,
    this.inputFormatters,
    this.enabled,
    double? cursorWidth,
    this.cursorRadius,
    this.cursorColor,
    this.keyboardAppearance,
    EdgeInsets? scrollPadding,
    bool? enableInteractiveSelection,
    this.selectionControls,
    this.onTap,
    this.onTapOutside,
    this.strutStyle,
    this.textAlignVertical,
    bool? readOnly,
    this.showCursor,
    String? obscuringCharacter,
    this.smartDashesType,
    this.smartQuotesType,
    bool? enableSuggestions,
    this.minLines,
    bool? expands,
    this.onAppPrivateCommand,
    this.cursorHeight,
    ui.BoxHeightStyle? selectionHeightStyle,
    ui.BoxWidthStyle? selectionWidthStyle,
    DragStartBehavior? dragStartBehavior,
    this.mouseCursor,
    this.buildCounter,
    this.scrollController,
    this.scrollPhysics,
    this.autofillHints,
    this.clipBehavior = Clip.hardEdge,
    this.restorationId,
    this.scribbleEnabled = true,
    this.enableIMEPersonalizedLearning = true,
    this.contextMenuBuilder = _defaultContextMenuBuilder,
    this.spellCheckConfiguration,
    this.magnifierConfiguration,
    CustomButton? customButton,
    CustomButtonDefaultAction? customButtonDefaultAction,
    this.customButtonAction,
    this.customButtonColor,
    this.customButtonIcon,
    double? customButtonIconSize,
    double? customButtonIconPadding,
    bool? clearErrorOnTextChange = true,
  })  : decoration = decoration ?? const InputDecoration(),
        textCapitalization = textCapitalization ?? TextCapitalization.none,
        textAlign = textAlign ?? TextAlign.start,
        autofocus = autofocus ?? false,
        obscureText = obscureText ?? false,
        autocorrect = autocorrect ?? true,
        maxLines = maxLines ?? 1,
        cursorWidth = cursorWidth ?? 2.0,
        scrollPadding = scrollPadding ?? const EdgeInsets.all(20.0),
        enableInteractiveSelection = enableInteractiveSelection ?? true,
        readOnly = readOnly ?? false,
        obscuringCharacter = obscuringCharacter ?? '•',
        enableSuggestions = enableSuggestions ?? true,
        expands = expands ?? false,
        selectionHeightStyle = selectionHeightStyle ?? ui.BoxHeightStyle.tight,
        selectionWidthStyle = selectionWidthStyle ?? ui.BoxWidthStyle.tight,
        dragStartBehavior = dragStartBehavior ?? DragStartBehavior.start,
        customButton = customButton ?? CustomButton.NONE,
        customButtonDefaultAction =
            customButtonDefaultAction ?? CustomButtonDefaultAction.NONE,
        customButtonIconSize = customButtonIconSize ?? 24,
        customButtonIconPadding = customButtonIconPadding ?? 8,
        clearErrorOnTextChange = clearErrorOnTextChange ?? true;

  @override
  AdeptTextFieldState createState() => AdeptTextFieldState();
}

class AdeptTextFieldState extends State<AdeptTextField> {
  late final AdeptTextEditingController controller;
  bool hasText = false;
  bool hasError = false;
  String? lastText;
  late bool obscureText;
  late InputDecoration decoration;

  AdeptTextFieldState();

  @override
  void initState() {
    super.initState();
    controller = widget.controller ?? AdeptTextEditingController();
    if (widget.customButtonDefaultAction ==
        CustomButtonDefaultAction.PASSWORD_TOGGLE) {
      obscureText = true;
    } else {
      obscureText = widget.obscureText;
    }
    decoration = widget.decoration;
    hasText = controllerHasText;
    hasError = controller.hasError;
    lastText = controller.text;
    controller.addListener(onTextChanged);
  }

  void updateDecoration() {
    if (widget.customButton != CustomButton.NONE) {
      decoration = decoration.copyWith(
        errorText: controller.error,
        suffixIcon: !hasText
            ? const SizedBox()
            : IconButton(
                icon: getCustomButtonIcon,
                iconSize: widget.customButtonIconSize,
                padding: EdgeInsets.all(widget.customButtonIconPadding),
                onPressed: () {
                  if (widget.customButtonAction != null) {
                    widget.customButtonAction?.call();
                  } else {
                    switch (widget.customButtonDefaultAction) {
                      case CustomButtonDefaultAction.CLEAR:
                        controller.clear();
                        break;
                      case CustomButtonDefaultAction.PASSWORD_TOGGLE:
                        obscureText = !obscureText;
                        setState(() {});
                        break;
                      default:
                    }
                  }
                },
              ),
        suffixText: null,
      );
    }
  }

  bool get controllerHasText {
    bool show = false;
    if (widget.customButton == CustomButton.ALWAYS) show = true;
    if (widget.customButton == CustomButton.ON_TEXT &&
        controller.text.isNotEmpty) show = true;
    return show;
  }

  void onTextChanged() {
    bool controllerHasText = this.controllerHasText;
    if (controllerHasText != hasText) {
      hasText = controllerHasText;
      hasError = controller.hasError;
      setState(() {});
    } else if (hasError != controller.hasError) {
      hasError = controller.hasError;
      setState(() {});
    }
    if (widget.onChanged != null && lastText != controller.text) {
      widget.onChanged!(lastText = controller.text);
    }
    lastText = controller.text;
  }

  Widget get getCustomButtonIcon {
    if (widget.customButtonIcon != null) {
      return Icon(
        widget.customButtonIcon,
        color: widget.customButtonColor,
      );
    }
    switch (widget.customButtonDefaultAction) {
      case CustomButtonDefaultAction.CLEAR:
        return Icon(
          Icons.clear,
          color: widget.customButtonColor,
        );
      case CustomButtonDefaultAction.PASSWORD_TOGGLE:
        return Icon(
          obscureText ? Icons.visibility_off : Icons.visibility,
          color: widget.customButtonColor,
        );
      default:
        return const SizedBox();
    }
  }

  @override
  void didUpdateWidget(covariant AdeptTextField oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (oldWidget.controller != controller) {
      oldWidget.controller?.removeListener(onTextChanged);
      controller.addListener(onTextChanged);
    }
  }

  @override
  Widget build(BuildContext context) {
    updateDecoration();
    return Theme(
      data: Theme.of(context).copyWith(splashColor: Colors.transparent),
      child: TextField(
        key: widget.key,
        controller: controller,
        focusNode: widget.focusNode,
        decoration: decoration,
        keyboardType: widget.keyboardType,
        textInputAction: widget.textInputAction,
        textCapitalization: widget.textCapitalization,
        style: widget.style,
        strutStyle: widget.strutStyle,
        textAlign: widget.textAlign,
        textAlignVertical: widget.textAlignVertical,
        textDirection: widget.textDirection,
        readOnly: widget.readOnly,
        showCursor: widget.showCursor,
        autofocus: widget.autofocus,
        obscuringCharacter: widget.obscuringCharacter,
        obscureText: obscureText,
        autocorrect: widget.autocorrect,
        smartDashesType: widget.smartDashesType,
        smartQuotesType: widget.smartQuotesType,
        enableSuggestions: widget.enableSuggestions,
        maxLines: widget.maxLines,
        minLines: widget.minLines,
        expands: widget.expands,
        maxLength: widget.maxLength,
        maxLengthEnforcement: widget.maxLengthEnforcement,
//        onChanged: widget.onChanged, //onChanged function is managed above in the onTextChanged function
        onEditingComplete: widget.onEditingComplete,
        onSubmitted: widget.onSubmitted,
        onAppPrivateCommand: widget.onAppPrivateCommand,
        inputFormatters: widget.inputFormatters,
        enabled: widget.enabled,
        cursorWidth: widget.cursorWidth,
        cursorHeight: widget.cursorHeight,
        cursorRadius: widget.cursorRadius,
        cursorColor: widget.cursorColor,
        selectionHeightStyle: widget.selectionHeightStyle,
        selectionWidthStyle: widget.selectionWidthStyle,
        keyboardAppearance: widget.keyboardAppearance,
        scrollPadding: widget.scrollPadding,
        dragStartBehavior: widget.dragStartBehavior,
        enableInteractiveSelection: widget.enableInteractiveSelection,
        selectionControls: widget.selectionControls,
        onTap: widget.onTap,
        onTapOutside: widget.onTapOutside,
        mouseCursor: widget.mouseCursor,
        buildCounter: widget.buildCounter,
        scrollController: widget.scrollController,
        scrollPhysics: widget.scrollPhysics,
        autofillHints: widget.autofillHints,
        clipBehavior: widget.clipBehavior,
        restorationId: widget.restorationId,
        scribbleEnabled: widget.scribbleEnabled,
        enableIMEPersonalizedLearning: widget.enableIMEPersonalizedLearning,
        contextMenuBuilder: widget.contextMenuBuilder,
        spellCheckConfiguration: widget.spellCheckConfiguration,
        magnifierConfiguration: widget.magnifierConfiguration,
      ),
    );
  }

  @override
  void dispose() {
    controller.removeListener(onTextChanged);
    super.dispose();
  }
}
